# scorepgn.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide class which switches board display between game score and analysis.

The ScorePGN class is used to to display games and repertoires.

The games and repertoires are in Portable Game Notation (PGN) text.
"""


class ScorePGN:
    """Display current game or analysis position on the board.

    These methods display the game or repertoire PGN score, or the engine
    analysis for the PGN score.

    """

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def analysis_current_item(self, event=None):
        """Select current PGN score analysis."""
        del event
        if self._get_game_position_analysis():
            self.analysis.apply_colouring_to_variation_back_to_main_line()
        else:
            self.analysis.clear_current_range()
            self.analysis.clear_moves_played_in_variation_colouring_tag()
            self.analysis.current = None
        self.analysis.set_current()
        self.analysis.set_game_board()
        self.set_board_pointer_move_bindings(False)
        self.analysis.set_board_pointer_move_bindings(True)
        self.set_score_pointer_item_navigation_bindings(False)
        self.analysis.set_score_pointer_item_navigation_bindings(True)
        self.set_score_pointer_to_score_bindings(True)
        self.set_analysis_score_pointer_to_analysis_score_bindings(False)
        self.set_toggle_game_analysis_bindings(True)
        self.analysis.score.focus_set()
        self._set_game_position_analysis(True)
        self._init_takefocus_widget(self.analysis.score)

    def _current_pgn_score(self, cuiai):
        """Select current PGN score on display.

        Argument name follows history of separate functionally identical
        methods, all called current_item, originally present in gamedisplay
        and repertoiredisplay modules.

        """
        # cuiai should be referencing self given use of current_item() method,
        # but style of sibling *_item() methods is followed.
        # cuiai was cuigs in gamedisplay, and cuirs in repertoiredisplay,
        # modules originally.
        self.analysis.clear_current_range()
        self.analysis.clear_moves_played_in_variation_colouring_tag()
        if self.current is None:
            cuiai.board.set_board(cuiai.fen_tag_square_piece_map())
        else:
            cuiai.board.set_board(cuiai.tagpositionmap[self.current][0])
        cuiai.set_game_list()
        cuiai.analysis.set_board_pointer_move_bindings(False)
        cuiai.set_board_pointer_move_bindings(True)
        cuiai.analysis.set_score_pointer_item_navigation_bindings(False)
        cuiai.set_score_pointer_item_navigation_bindings(True)
        cuiai.set_score_pointer_to_score_bindings(False)
        cuiai.set_analysis_score_pointer_to_analysis_score_bindings(True)
        cuiai.set_toggle_game_analysis_bindings(True)
        self._init_takefocus_widget(self.score)
        cuiai.takefocus_widget.focus_set()



# ndbmdu_dir.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with ndbm.chessndbmdu to database.

The default segment size, 4000 bytes, leads to a
'HASH: Out of overflow pages.  Increase page size'
error when importing about 2000 games split about evenly between two PGN files.

The same import succeeds if chesstab.tools.create_chesstab_database is used to
create the database with a segment size of 16 bytes.  A more reasonable size,
500 bytes, the smallest intended for production may work on the example tried
too.
"""


if __name__ == "__main__":

    from .directory_widget import DirectoryWidget
    from ..ndbm.chessndbmdu import chess_database_du

    DirectoryWidget(chess_database_du, "dbm.ndbm")



# cqltext.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display a Chess Query Language (ChessQL) statement.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

"""
# The previous CQL syntax partially supported was version 5.1 found at:
# https://web.archive.org/web/20140130143815/http://www.rbnn.com/cql/
# (www.rbnn.com is no longer availabable).

import tkinter
import tkinter.messagebox

from ..core.cqlstatement import CQLStatement
from .blanktext import NonTagBind, BlankText
from .sharedtext import SharedText, SharedTextEngineText, SharedTextScore


class CQLTextListGamesError(Exception):
    """Exception class for display of lists of games."""


class CQLText(SharedText, SharedTextEngineText, SharedTextScore, BlankText):
    """ChessQL statement widget.

    panel is used as the panel argument for the super().__init__ call.

    ui is the user interface manager for an instance of CQLText, usually an
    instance of ChessUI.

    items_manager is used as the items_manager argument for the
    super().__init__ call.

    itemgrid is the ui reference to the DataGrid from which the record was
    selected.

    Subclasses are responsible for providing a geometry manager.

    Attribute _most_recent_bindings is set to indicate the initial set of
    event bindings.  Instances will override this as required.

    """

    def __init__(
        self, panel, ui=None, items_manager=None, itemgrid=None, **ka
    ):
        """Create widgets to display ChessQL statement."""
        super().__init__(panel, items_manager=items_manager, **ka)
        self.ui = ui
        self.itemgrid = itemgrid

        # Selection rule parser instance to process text.
        self.cql_statement = CQLStatement()
        # Not sure this is needed or wanted.
        # self.cql_statement.dbset = ui.base_games.datasource.dbset

    def set_and_tag_item_text(self, reset_undo=False):
        """Display the ChessQL statement as text.

        reset_undo causes the undo redo stack to be cleared if True.  Set True
        on first display of a ChessQL statement for editing so that repeated
        Ctrl-Z in text editing mode recovers the original ChessQL statement.

        """
        if not self._is_text_editable:
            self.score.configure(state=tkinter.NORMAL)
        self.score.delete("1.0", tkinter.END)
        self._map_cql_statement()
        if self._most_recent_bindings != NonTagBind.NO_EDITABLE_TAGS:
            self._bind_for_primary_activity()
        if not self._is_text_editable:
            self.score.configure(state=tkinter.DISABLED)
        if reset_undo:
            self.score.edit_reset()

    def set_statusbar_text(self):
        """Set status bar to display ChessQL statement name."""
        self.ui.statusbar.set_status_text(self.cql_statement.get_name_text())

    def get_name_cql_statement_text(self):
        """Return text from CQL statement Text widget."""
        text = self.score.get("1.0", tkinter.END).strip()
        return text

    def _map_cql_statement(self):
        """Convert tokens to text and show in CQL statement Text widget."""
        # No mapping of tokens to text in widget (yet).
        self.score.insert(
            tkinter.INSERT, self.cql_statement.get_name_statement_text()
        )

    def _get_partial_key_cql_statement(self):
        """Return ChessQL statement for use as partial key."""
        if self.cql_statement.is_statement():

            # Things must be arranged so a tuple, not a list, can be returned.
            # return tuple(self.cql_statement.position)
            return self.cql_statement.get_statement_text()  # Maybe!

        return False

    def refresh_game_list(self):
        """Display games with position matching selected ChessQL statement."""
        grid = self.itemgrid
        if grid is None:
            return
        if grid.get_database() is None:
            return
        cqls = self.cql_statement
        if cqls.cql_error:
            grid.datasource.get_cql_statement_games(None, None)
        else:
            try:
                if self._is_text_editable:
                    grid.datasource.get_cql_statement_games(cqls, None)
                else:
                    grid.datasource.get_cql_statement_games(
                        cqls, self.recalculate_after_edit
                    )
            except AttributeError as exc:
                if str(exc) == "'NoneType' object has no attribute 'answer'":
                    msg = "".join(
                        (
                            "Unable to list games for ChessQL statement, ",
                            "probably because an 'empty square' is in the ",
                            "query (eg '.a2-3'):\n\n",
                            "The reported  error is:\n\n",
                            str(exc),
                        )
                    )
                else:
                    msg = "".join(
                        (
                            "Unable to list games for ChessQL statement:\n\n",
                            "The reported error is:\n\n",
                            str(exc),
                        )
                    )
                grid.datasource.get_cql_statement_games(None, None)
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement",
                    message=msg,
                )
        grid.partial = self._get_partial_key_cql_statement()
        # grid.rows = 1
        grid.load_new_index()

        # Get rid of the 'Please wait ...' status text.
        self.ui.statusbar.set_status_text()

        if cqls.cql_error:
            if self.ui.database is None:
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement Error",
                    message=cqls.cql_error.get_error_report(),
                )
            else:
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement Error",
                    message=cqls.cql_error.add_error_report_to_message(
                        ("An empty game list will be displayed.")
                    ),
                )
        elif grid.datasource.not_implemented:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="ChessQL Statement Not Implemented",
                message="".join(
                    (
                        "These filters are not implemented and ",
                        "are ignored:\n\n",
                        "\n".join(sorted(grid.datasource.not_implemented)),
                    )
                ),
            )



# toplevelpgn.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which switch between game score and analysis display."""

# These classes existed in the gamedbdelete, gamedbdisplay, gamedbedit,
# gamedbshow, repertoiredbdelete, repertoiredbdisplay, repertoiredbedit,
# and repertoiredbshow, modules which display Portable Game Notation (PGN)
# text.
# All methods in the classes in this module existed as multiple copies in the
# modules named above.  They are now deleted from those modules.
# The classes in this module represent the different sets of classes with
# methods in common.
# The ToplevelPGN class is populated with the methods identical in
# GameToplevel, RepertoireToplevel, GameToplevelEdit, and
# RepertoireToplevelEdit, which were then removed from those two classes.

import tkinter.messagebox

from .eventspec import EventSpec
from .scorepgn import ScorePGN
from .topleveltext import DeleteText, EditText, ShowText


class ToplevelPGN(ScorePGN):
    """Switch focus between game score and analysis for a game.

    The game and repertoire display methods assume there is an associated list
    of games for the current position.  The *Toplevel* classes do not have
    this list, so the method which updates it is overridden.

    """

    binding_labels = (
        EventSpec.analysis_to_scoresheet,
        EventSpec.scoresheet_to_analysis,
    )

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def set_game_list(self):
        """Display list of records in grid.

        Called after each navigation event on a PGN score including switching
        from one PGN score to another.

        """
        # Score.set_game_list() expects instance to have itemgrid attribute
        # bound to a DataGrid subclass instance, but the Dialogue* instances
        # can live without this being present.
        # It may be more appropriate to override set_game_list to do nothing so
        # there is a way of deleting a record without tracking games containing
        # the same positions.
        try:
            super().set_game_list()
        except AttributeError:
            if self.itemgrid is not None:
                raise

    def generate_popup_navigation_maps(self):
        """Return tuple of widget navigation map and switch to analysis map."""
        navigation_map = {}
        local_map = {
            EventSpec.scoresheet_to_analysis: self.analysis_current_item,
        }
        return navigation_map, local_map

    def current_item(self, event=None):
        """Select current PGN score on display."""
        del event
        self._current_pgn_score(self)


class _ToplevelPGN:
    """Provide methods shared by this module's public _ToplevelPGN subclasses.

    The subclasses are DeletePGN, EditPGN, and ShowPGN.
    """

    def _initialize_item_bindings(self, item):
        """Initialize keypress and buttonpress bindings for item."""
        self.bind_buttons_to_widget(item.score)
        self.bind_buttons_to_widget(item.analysis.score)
        item.set_score_pointer_to_score_bindings(False)
        item.set_analysis_score_pointer_to_analysis_score_bindings(True)
        item.set_toggle_game_analysis_bindings(True)
        item.set_board_pointer_move_bindings(True)


class ShowPGNToplevel(_ToplevelPGN, ShowText):
    """Show original PGN."""


class DeletePGNToplevel(_ToplevelPGN, DeleteText):
    """Show original PGN for record deletion."""


class EditPGNToplevel(_ToplevelPGN, EditText):
    """Show original and editable PGN versions for record editing."""

    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        This method extends the version in EditText superclass.

        """
        text = self.newview.get_score_error_escapes_removed()
        self.newobject.value.load(repr(text))
        if not self.newobject.value.collected_game.is_pgn_valid():
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.parent,
                title="".join(("Edit ", self.pgn_score_name)),
                message=self.pgn_score_name.lower().join(
                    (
                        "The edited ",
                        "".join(
                            (
                                " contains at least one illegal move in PGN.",
                                "\n\nPlease re-confirm request to edit ",
                            )
                        ),
                        ".",
                    )
                ),
            ):
                return False
            self.newobject.value.set_game_source(self.pgn_score_source)
        return super().dialog_ok()



# positionrow.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets to display tag roster details of games matching position."""
# Put transposition moves in column 0 rather than column 4.
# Display just the moves played in, and to reach, the position.

import tkinter
from ast import literal_eval

from solentware_grid.gui.datarow import (
    DataRow,
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from . import constants
from ..core.chessrecord import ChessDBrecordGamePosition
from .positionscore import PositionScore
from ..shared.allrow import AllRow
from ..shared.game_position import GamePosition


class ChessDBrowPosition(
    GamePosition, AllRow, ChessDBrecordGamePosition, DataRow
):
    """Define row in list of games for given position.

    Add row methods to the chess game record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text="Transposition",
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="move"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super().__init__()
        self.ui = ui
        self.set_database(database)
        self.score = None
        self.row_specification = [
            {
                WIDGET: tkinter.Text,
                WIDGET_CONFIGURE: dict(
                    height=0,
                    relief=tkinter.FLAT,
                    font=constants.LISTS_OF_GAMES_FONT,
                    wrap=tkinter.NONE,
                    borderwidth=2,  # fill cell to row height from labels
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def grid_row(self, textitems=(), **kargs):
        """Delegate to __grid_row and reurn response."""
        return self.__grid_row(textitems=textitems, **kargs)

    def __grid_row(
        self, textitems=(), position=None, context=(None, None, None), **kargs
    ):
        """Set textitems to record value, delegate, return response.

        Create textitems argument for ChessDBrowPosition instance.

        The record value is expected to contain transposition details.

        textitems arguments is ignored and is present for compatibility.

        """
        del textitems
        del position
        self.row_specification[0][WIDGET_CONFIGURE]["context"] = context
        return super().grid_row(
            textitems=(literal_eval(self.srvalue),), **kargs
        )

    def set_background(self, widgets, background):
        """Set background colour of widgets.

        widgets - list((widget, specification), ...).
        background - the background colour.

        Each element of widgets will have been created by make_row_widgets()
        or DataHeader.make_header_widgets() and reused by DataGrid instance
        in a data row.

        """
        for widget, rowspec in zip(widgets, self.row_specification):
            if "background" not in rowspec[WIDGET_CONFIGURE]:
                widget[0].configure(background=background)

    def populate_widget(self, widget, cnf=None, text=None, **kw):
        """Delegate to __populate_widget."""
        self.__populate_widget(widget, cnf=cnf, text=text, **kw)

    def __populate_widget(
        self, widget, cnf=None, text=None, context=None, **kw
    ):
        """Delegate for tkinter.Label widget, put text in a PositionScore.

        The context argument is only relevant for the PositionScore.

        """
        if isinstance(widget, tkinter.Label):
            super().populate_widget(widget, cnf=cnf, text=text, **kw)
            return
        # This is the place to implement a pool of pre-processed PositionScore
        # instances which only need to call the colour_score() method rather
        # than the process_score() method.
        # Goal is to speed up populating widget showing games containing
        # current position of active game when the widget is able to list more
        # than a few (<5 say) games.
        if text:
            if self.score is None:
                self.score = PositionScore(widget, ui=self.ui, **kw)
            self.score.process_score(text=text, context=context)
        kw["width"] = self.score.score.count("1.0", tkinter.END, "chars")[0]
        widget.configure(cnf=cnf, **kw)


def chess_db_row_position(chessui):
    """Return function that returns ChessDBrowPosition instance for chessui.

    chessui is a chess_ui.ChessUI instance.

    The returned function takes a ChessDatabase instance as it's argument.
    """

    def make_position(database=None):
        return ChessDBrowPosition(database=database, ui=chessui)

    return make_position



# archivedu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The Archivedu class for interfaces to single file databases.

Most of the interface modules that can use the Litedu module have 'lite' in
their name.

This module is relevant to the apsw and sqlite3 interfaces to Sqlite3, and
to the gnu, ndbm, unqlite, and vedis, interfaces to their respective
'key:value' databases.

"""

import os
import bz2


class Archivedu:
    """Provide deferred update archive methods shared by various interfaces.

    All the supported engines put the whole database in a single file so
    can use the same methods to manage temporary backups which may exist
    while opening and checking the database.
    """

    def archive(self, flag=None, names=None):
        """Write a bz2 backup of file containing games.

        Intended to be a backup in case import fails.

        """
        if names is None:
            return False
        if not self.delete_archive(flag=flag, names=names):
            return None
        if flag:
            _archive(names)
        return True

    @staticmethod
    def delete_archive(flag=None, names=None):
        """Delete a bz2 backup of file containing games."""
        if names is None:
            return False
        if flag:
            _delete_archive(names)
        return True


# Snippet needed in dbdu.Dbdu.archive() method too.
def _archive(names):
    for name in names:
        compressor = bz2.BZ2Compressor()
        archiveguard = ".".join((name, "grd"))
        archivename = ".".join((name, "bz2"))
        with open(name, "rb") as file_in, open(archivename, "wb") as file_out:
            inp = file_in.read(10000000)
            while inp:
                compressed = compressor.compress(inp)
                if compressed:
                    file_out.write(compressed)
                inp = file_in.read(10000000)
            compressed = compressor.flush()
            if compressed:
                file_out.write(compressed)
        with open(archiveguard, "wb"):
            pass


# Snippet needed in dbdu.Dbdu.delete_archive() method too.
def _delete_archive(names):
    for name in names:
        archiveguard = ".".join((name, "grd"))
        archivename = ".".join((name, "bz2"))
        try:
            os.remove(archiveguard)
        except FileNotFoundError:
            pass
        try:
            os.remove(archivename)
        except FileNotFoundError:
            pass



# repertoiredbdelete.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete toplevel to delete repertoire record."""

from solentware_grid.gui.datadelete import DataDelete

from pgn_read.core.parser import PGN

from ..core.constants import TAG_OPENING
from .repertoiretoplevel import RepertoireToplevel
from .toplevelpgn import DeletePGNToplevel


class RepertoireDbDelete(DeletePGNToplevel, DataDelete):
    """Delete PGN text for repertoire from database.

    parent is used as the master argument in a RepertoireToplevel call.

    ui is used as the ui argument in a RepertoireToplevel call.

    parent, oldobject, and the RepertoireToplevel instance created, are used
    as arguments in the super.__init__ call.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    pgn_score_name = "Repertoire"

    def __init__(self, parent=None, oldobject=None, ui=None):
        """Extend and create toplevel widget for deleting chess game."""
        # Toplevel title set '' in __init__ and to proper value in _initialize.
        super().__init__(
            instance=oldobject,
            parent=parent,
            oldview=RepertoireToplevel(master=parent, ui=ui),
            title="",
        )
        self._initialize()

    @property
    def ui_base_table(self):
        """Return the User Interface RepertoireGrid object."""
        return self.ui.base_repertoires

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.games_and_repertoires_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'read-only' view."""
        return self.oldview.ui

    def _set_item(self, view, object_):
        """Populate view with the repertoire extracted from object_."""
        self._set_default_source_for_object(object_)
        view.set_position_analysis_data_source()
        view.collected_game = next(
            PGN(game_class=view.gameclass).read_games(object_.get_srvalue())
        )
        view.set_and_tag_item_text()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a Repertoire object_.

        Default value of object_ is object attribute from DataDelete class.

        """
        if object_ is None:
            object_ = self.object
        try:
            return "  ".join(
                (
                    self.pgn_score_name.join(("Delete ", ":")),
                    object_.value.collected_game.pgn_tags[TAG_OPENING],
                )
            )
        except TypeError:
            return self.pgn_score_name.join(
                ("Delete ", " - name unknown or invalid")
            )
        except KeyError:
            return self.pgn_score_name.join(
                ("Delete ", " - name unknown or invalid")
            )

    def _set_default_source_for_object(self, object_=None):
        """Set default source for Toplevel containing a Repertoire object_.

        Default value of object_ is object attribute from DataDelete class.

        """



# repertoirerow.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets that display tag roster details of repertoire."""

import tkinter

from solentware_grid.gui.datarow import (
    DataRow,
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from pgn_read.core.constants import TAG_RESULT

from ..core.chessrecord import ChessDBrecordRepertoireTags
from .repertoiredbedit import RepertoireDbEdit
from .repertoiredbdelete import RepertoireDbDelete
from .repertoiredbshow import RepertoireDbShow
from . import constants
from ..core.constants import TAG_OPENING, REPERTOIRE_GAME_TAGS
from ..shared.allrow import AllRow


class ChessDBrowRepertoire(AllRow, ChessDBrecordRepertoireTags, DataRow):
    """Define row in list of repertoires.

    Add row methods to the chess game record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_OPENING,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="player"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_RESULT,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=1, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="result"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super().__init__()
        self.ui = ui
        self.set_database(database)
        self.row_specification = [
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=1, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def show_row(self, dialog, oldobject):
        """Return a RepertoireDbShow toplevel for instance.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return RepertoireDbShow(parent=dialog, oldobject=oldobject, ui=self.ui)

    def delete_row(self, dialog, oldobject):
        """Return a RepertoireDbDelete toplevel for instance.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return RepertoireDbDelete(
            parent=dialog, oldobject=oldobject, ui=self.ui
        )

    def edit_row(self, dialog, newobject, oldobject, showinitial=True):
        """Return a RepertoireDbEdit toplevel for instance.

        dialog - a Toplevel
        newobject - a ChessDBrecordGame containing original data to be edited
        oldobject - a ChessDBrecordGame containing original data
        showintial == True - show both original and edited data

        """
        return RepertoireDbEdit(
            newobject=newobject,
            parent=dialog,
            oldobject=oldobject,
            showinitial=showinitial,
            ui=self.ui,
        )

    def grid_row(self, textitems=(), **kargs):
        """Set textitems to repertoire name, delegate, return response.

        Create textitems argument for ChessDBrowRepertoire instance.

        textitems arguments is ignored and is present for compatibility.

        """
        tags = self.value.collected_game.pgn_tags
        return super().grid_row(
            textitems=(
                tags.get(TAG_OPENING, "?"),
                tags.get(TAG_RESULT, "?"),
            ),
            **kargs
        )

    def _get_tags_display_order(self, pgn):
        """Return Tags not given their own column in display order."""
        del pgn
        tag_values = []
        tags = self.value.collected_game.pgn_tags
        for item in sorted(tags.items()):
            if item[0] not in REPERTOIRE_GAME_TAGS:
                tag_values.append(item)
        return tag_values


def chess_db_row_repertoire(chessui):
    """Return function that returns ChessDBrowRepertoire instance for chessui.

    chessui is a chess_ui.ChessUI instance.

    The returned function takes a ChessDatabase instance as it's argument.
    """

    def make_position(database=None):
        return ChessDBrowRepertoire(database=database, ui=chessui)

    return make_position



# export_selection_rule.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess selection rule exporters."""

from . import chessrecord, filespec


def export_selected_selection_rules(grid, filename):
    """Export selected selection rule statements to textfile."""
    if filename is None:
        return
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        instance = chessrecord.ChessDBrecordQuery()
        instance.set_database(database)
        with open(filename, "w") as gamesout:
            for bookmark in sorted(grid.bookmarks):
                instance.load_record(
                    database.get_primary_record(
                        filespec.SELECTION_FILE_DEF, bookmark[0]
                    )
                )
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
        return
    database = grid.get_data_source().dbhome
    instance = chessrecord.ChessDBrecordQuery()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.SELECTION_FILE_DEF, filespec.SELECTION_FILE_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return



# pgn.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Subclasses of pgn_read.core.game.Game class used by ChessTab.

Game* classes replace PGN* classes in ChessTab version 4.3.

"""
from pgn_read.core.game import Game, GameIndicateCheck
from pgn_read.core.squares import Squares
from pgn_read.core.constants import (
    FEN_WHITE_KING,
    FEN_WHITE_QUEEN,
    FEN_WHITE_ROOK,
    FEN_WHITE_BISHOP,
    FEN_WHITE_KNIGHT,
    FEN_WHITE_PAWN,
    FEN_BLACK_KING,
    FEN_BLACK_QUEEN,
    FEN_BLACK_ROOK,
    FEN_BLACK_BISHOP,
    FEN_BLACK_KNIGHT,
    FEN_BLACK_PAWN,
    FEN_WHITE_ACTIVE,
)

from chessql.core.constants import ANY_WHITE_PIECE_NAME, ANY_BLACK_PIECE_NAME

from .constants import (
    TAG_OPENING,
    REPERTOIRE_TAG_ORDER,
    REPERTOIRE_GAME_TAGS,
    MOVE_NUMBER_KEYS,
)

MAP_PGN_PIECE_TO_CQL_COMPOSITE_PIECE = {
    FEN_WHITE_KING: ANY_WHITE_PIECE_NAME,
    FEN_WHITE_QUEEN: ANY_WHITE_PIECE_NAME,
    FEN_WHITE_ROOK: ANY_WHITE_PIECE_NAME,
    FEN_WHITE_BISHOP: ANY_WHITE_PIECE_NAME,
    FEN_WHITE_KNIGHT: ANY_WHITE_PIECE_NAME,
    FEN_WHITE_PAWN: ANY_WHITE_PIECE_NAME,
    FEN_BLACK_KING: ANY_BLACK_PIECE_NAME,
    FEN_BLACK_QUEEN: ANY_BLACK_PIECE_NAME,
    FEN_BLACK_ROOK: ANY_BLACK_PIECE_NAME,
    FEN_BLACK_BISHOP: ANY_BLACK_PIECE_NAME,
    FEN_BLACK_KNIGHT: ANY_BLACK_PIECE_NAME,
    FEN_BLACK_PAWN: ANY_BLACK_PIECE_NAME,
}


class GameDisplayMoves(GameIndicateCheck):
    """Add structures to support display of PGN moves."""

    def __init__(self):
        """Delegate then initialize moves to empty list."""
        super().__init__()
        self.moves = []

    # Replaces self.set_position_fen(self, fen=None)
    def set_initial_board_state(self, position_delta):
        """Initialize PGN score parser with Forsyth Edwards Notation position.

        fen defaults to the starting position for a game of chess.

        """
        super().set_initial_board_state(position_delta)
        if self._ravstack:
            self.moves = [(None, self._ravstack[-1][-1])]

    def modify_board_state(self, position_delta):
        """Delegate to superclass then append entry to moves."""
        super().modify_board_state(position_delta)
        self.moves.append((self._text[-1], self._ravstack[-1][-1]))


class GameMove(Game):
    """Generate data structures to verify moves returned from Chess Engines.

    set_initial_position() is not called in Game until the first movetext
    token is processed, but conversion of moves returned from chess engines
    to PGN needs the data structures set by the set_initial_position() call.

    GameMove allows this call to be done before processing the first movetext
    and supresses the call usually done with first movetext.

    """

    def __init__(self):
        """Delegate then set 'set initial position state' to None."""
        super().__init__()
        self._set_initial_position_state = None

    def set_initial_position(self):
        """Return 'set initial position state' after delegate to initialize."""
        if self._set_initial_position_state is None:
            self._set_initial_position_state = super().set_initial_position()
        return self._set_initial_position_state


class GameAnalysis(GameDisplayMoves):
    """Generate data to display Chess Engine analysis without ability to edit.

    The notion of mandatory PGN tags, like the 'seven tag roster', is removed
    from the GameDisplayMoves class.  Section 8.1.1: Seven Tag Roster of the
    PGN specification starts: 'There is a set of tags defined for mandatory
    use for archival storage of PGN data.  This is the STR (Seven Tag Roster)'.
    Thus the absence of all the STR tags does not prevent data from being
    called PGN data.  In particular the Result tag is not present fitting the
    Game Termination Marker '*' assigned to all analysis.

    Subclasses may use the PGN tag structure to manage analysis, but exactly
    what tags are defined is up to them.

    A single main move is required; to which chess engine analysis is attached
    as a sequence of RAVs, each RAV corresponding to one PV in a PV or multiPV
    response from a chess engine.

    """

    def is_tag_roster_valid(self):
        """Return True if the game's tag roster is valid."""
        tags = self._tags
        for value in tags.values():
            if len(value) == 0:
                # Tag value must not be null
                return False
        return True


def get_position_string(
    board,
    active_color,
    castling_availability,
    en_passant_target_square,
    halfmove_clock,
    fullmove_number,
):
    """Return position string for description of board.

    Format of position string is (I say bytes even though a str is returned):
    8 bytes with each set bit representing an occupied square from a8 to h1.
    n bytes corresponding to n set bits for occupied squares naming the piece
    on the square.
    1 byte value 'w' or 'b' naming side to move.
    1 or 2 bytes naming the square to which a pawn may move by capturing en
    passant. The only 1 byte value allowed is '-' meaning no en passant capture
    is possible.  The allowed 2 byte values are 'a6' to 'h6' and 'a3' to 'h3'.
    1 to 4 bytes indicating the castling moves possible if the appropriate side
    has the move.  Thus 'KQkq' means all four castling moves are possible and
    '-' indicates no casting moves are possible.  The other allowed value are
    obtained by removing one or more bytes from the original 'KQkq' without
    changing the order of those remaining.

    These values are intended as keys in an index and this structure puts keys
    for similar positions, specifically for castling and en passant
    differences, near each other.  En passant is before castling because the
    meaning of 'b', 'Q', and 'q', can be decided without using the 8 byte bit
    pattern: piece name or en passant or castling or whose move.

    """
    del halfmove_clock, fullmove_number
    squares = Squares.squares
    for square, piece in board.items():
        piece.set_square(square)
    return (
        sum(squares[s].bit for s in board)
        .to_bytes(8, "big")
        .decode("iso-8859-1")
        + "".join(p.name for p in sorted(board.values()))
        + active_color
        + en_passant_target_square
        + castling_availability
    )


# The classes and functions above this comment are used in, or via, the Game,
# GameEdit, and Score, classes.  Those below this comment are used in, or via,
# the Repertoire, RepertoireEdit, and database update, classes only.  The
# order has been changed to ease a potential split in ChessTab to provide a
# 'ChessView' subset capable of viewing PGN files without the database update
# options.


class GameRepertoireDisplayMoves(GameDisplayMoves):
    """Generate data to display a repertoire without ability to edit.

    The Seven Tag Roster is ignored, except for the Result tag, and a private
    Opening tag is mandatory instead.

    Export methods are provided for repertoires.

    """

    def is_tag_roster_valid(self):
        """Return True if the game's tag roster is valid."""
        tags = self._tags
        for value in tags.values():
            if len(value) == 0:
                # Tag value must not be null
                return False
        if TAG_OPENING not in tags:
            # A mandatory tag is missing
            return False
        return True

    def get_repertoire_pgn(self):
        """Return export format PGN for repertoire."""
        tags = self._tags
        tokens = []
        for tag in REPERTOIRE_TAG_ORDER:
            tokens.extend(
                [
                    "[",
                    tag,
                    ' "',
                    tags.get(tag, REPERTOIRE_GAME_TAGS[tag]),
                    '"]\n',
                ]
            )
        for tag, value in sorted(
            [tv for tv in tags.items() if tv[0] not in REPERTOIRE_GAME_TAGS]
        ):
            tokens.extend(["[", tag, ' "', value, '"]\n'])
        tokens.append(self.get_all_movetext_in_pgn_export_format())
        return "".join(tokens)

    def get_repertoire_pgn_no_comments(self):
        """Return export format PGN for repertoire excluding comments."""
        tags = self._tags
        tokens = []
        for tag in REPERTOIRE_TAG_ORDER:
            tokens.extend(
                [
                    "[",
                    tag,
                    ' "',
                    tags.get(tag, REPERTOIRE_GAME_TAGS[tag]),
                    '"]\n',
                ]
            )
        for tag, value in sorted(
            [tv for tv in tags.items() if tv[0] not in REPERTOIRE_GAME_TAGS]
        ):
            tokens.extend(["[", tag, ' "', value, '"]\n'])
        tokens.append(
            self.get_movetext_without_comments_in_pgn_export_format()
        )
        return "".join(tokens)


class GameRepertoireUpdate(Game):
    """Generate data structures to Update a repertoire on a database."""

    def is_tag_roster_valid(self):
        """Return True if the repertoire's tag roster is valid."""
        tags = self._tags
        for value in tags.values():
            if len(value) == 0:
                # Tag value must not be null
                return False
        if TAG_OPENING not in tags:
            # A mandatory tag is missing
            return False
        return True


class GameTags(Game):
    """Generate data structures to display the PGN Tags of a game.

    Comments on two methods are worth making:

    _disambiguate_move is defined but the only way it should be reached has
    been disabled by overriding _collecting_movetext.  The definition ensures
    PGN._disambiguate_move is never called.

    PGN._collecting_non_whitespace_while_searching is correct in this class
    too.

    """

    def append_piece_move(self, match):
        """Ignore piece move token and update board state with null.

        Tokens are assumed to represent legal moves.

        """
        if self._movetext_offset is None:
            self._ravstack.append(None)
            self._movetext_offset = len(self._text)
            self._text.append("")

    def append_pawn_move(self, match):
        """Ignore pawn move token and update board state with null.

        Pawn promotion moves are handled by append_pawn_promote_move.

        Tokens are assumed to represent legal moves.

        """
        if self._movetext_offset is None:
            self._ravstack.append(None)
            self._movetext_offset = len(self._text)
            self._text.append("")

    def append_pawn_promote_move(self, match):
        """Ignore pawn promotion move token and update board state with null.

        Tokens are assumed to represent legal moves.

        """
        if self._movetext_offset is None:
            self._ravstack.append(None)
            self._movetext_offset = len(self._text)
            self._text.append("")

    def append_castles(self, match):
        """Ignore castling move token and update board state with null.

        Tokens are assumed to represent legal moves.

        """
        if self._movetext_offset is None:
            self._ravstack.append(None)
            self._movetext_offset = len(self._text)
            self._text.append("")

    def append_game_termination(self, match):
        """Append game termination token to game score and update game state.

        Tokens are assumed to represent legal moves.

        """
        if self._movetext_offset is None:
            self._movetext_offset = len(self._text)
        self._text.append(match.group())
        try:
            self.repeat_board_state(self._position_deltas[-1])
        except IndexError:
            self.add_board_state_none(None)

    def append_token(self, match):
        """Ignore valid non-tag token which does not change board state.

        The game position for this token is the same as for the adjacent token,
        and the game state is adjusted to fit.

        """
        try:
            self.repeat_board_state(self._position_deltas[-1])
        except IndexError:
            self.add_board_state_none(None)

    append_reserved = append_token

    def append_start_rav(self, match):
        """Append start RAV token to game score and update board state.

        Put game in error state if a variation cannot be put at current place
        in game score.

        """
        if self._movetext_offset is None:
            self.append_token_and_set_error(match)
            return
        self.reset_board_state(None)
        self._ravstack.append(None)
        self._text.append(match.group())
        self._state_stack.append(self._state)

    def append_end_rav(self, match):
        """Append end RAV token to game score and update board state.

        Put game in error state if a variation cannot be finished at current
        place in game score.

        """
        if self._state is not None or self._movetext_offset is None:
            self.append_token_and_set_error(match)
            return None

        if self._movetext_offset is None:
            self.append_token_and_set_error(match)
            return None
        if len(self._ravstack) == 1:
            self.append_token_and_set_error(match)
            return None
        del self._ravstack[-1]
        del self._state_stack[-1]
        self._state = self._state_stack[-1]

        self.set_board_state(None)
        self._text.append(match.group())
        return True

    def append_other_or_disambiguation_pgn(self, match):
        """Ignore token.

        'Qb3c2' can mean move the queen on b3 to c2 when all whitespace is
        removed from PGN movetext.  This method processes the 'c2' when it is
        consumed from the input: 'c2' is processed by peeking at the input when
        processing the 'Qb3'.

        """

    def append_token_after_error(self, match):
        """Ignore token after an error has been found."""

    def append_game_termination_after_error(self, match):
        """Append game termination token to game score.

        Tokens are assumed to represent legal moves.

        """
        self._text.append(match.group())

    def append_start_rav_after_error(self, match):
        """Append start RAV marker to game score and prepare error recovery.

        Rest of RAV is assumed to be in error, but after next matching end
        RAV marker it is worth assuming it possible to pick up from the
        valid position just before this start RAV marker.

        """
        self._text.append(match.group())
        self._state_stack.append(self._state)

    def append_end_rav_after_error(self, match):
        """Append end RAV marker to game score and reset position to pre-RAV.

        It is assumed the pre-RAV position is valid (otherwise the error
        state would have begun earlier).

        """
        if len(self._state_stack) > 1:
            if self._state_stack[-2] == self._state_stack[-1]:
                self._text.append(match.group())
                del self._state_stack[-1]
                self._state = self._state_stack[-1]
            else:

                # Cannot call append_end_rav() method because it tests some
                # conditions that should be true when errors are absent.
                if self._movetext_offset is None:
                    self.append_token_and_set_error(match)
                    return
                if len(self._ravstack) == 1:
                    self.append_token_and_set_error(match)
                    return
                del self._ravstack[-1]
                del self._state_stack[-1]
                self._state = self._state_stack[-1]
                self.set_board_state(None)
                self._text.append(match.group())

        else:
            self._text.append(match.group())


class GameRepertoireTags(GameTags):
    """Generate data structures to display the PGN Tags of a repertoire.

    The notion of mandatory PGN tags, like the 'seven tag roster', means the
    TAG_OPENING tag.

    """

    def is_tag_roster_valid(self):
        """Return True if the game's tag roster is valid."""
        tags = self._tags
        for value in tags.values():
            if len(value) == 0:
                # Tag value must not be null
                return False
        if TAG_OPENING not in tags:
            # A mandatory tag is missing
            return False
        return True


class GameUpdate(Game):
    """Prepare indicies after each token has been processed."""

    # self.positions, and the three similar, renamed to self.positionkeys.
    # self.movenumber is changed to self.halfmovenumber.
    def __init__(self):
        """Delegate then prepare to collect positions and piece locations."""
        super().__init__()
        self.positionkeys = []
        self.piecesquaremovekeys = []
        self.piecemovekeys = []
        self.squaremovekeys = []
        self.halfmovenumber = None
        self.variationnumber = None
        self.currentvariation = None
        self._variation = None

    # Replaces self.set_position_fen(self, fen=None)
    def set_initial_board_state(self, position_delta):
        """Initialize PGN score parser with Forsyth Edwards Notation position.

        fen defaults to the starting position for a game of chess.

        """
        super().set_initial_board_state(position_delta)
        if self._active_color == FEN_WHITE_ACTIVE:
            self.halfmovenumber = [(self._fullmove_number - 1) * 2]
        else:
            self.halfmovenumber = [self._fullmove_number * 2 - 1]
        self.variationnumber = [0]
        self._variation = "".join(
            _convert_integer_to_length_hex(i) for i in self.variationnumber
        )

    def reset_board_state(self, position_delta):
        """Delegate then append variation number to fit ravstack level."""
        super().reset_board_state(position_delta)
        if len(self._ravstack) > len(self.variationnumber):
            self.variationnumber.append(0)

    # Why '[len(self._ravstack)-1]' rather than '[-1]'?  The '[len()]' version
    # came from PGNUpdate in chesstab, without the -1 adjustment, and seems to
    # work.
    def set_board_state(self, position_delta):
        """Delegate then increment variation number.

        Both numeric and string variation numbers are kept in step in case
        there is another variation at this level.  For example:
        '... Ba7 ) ( Nf4 ...'.
        """
        super().set_board_state(position_delta)
        self.variationnumber[len(self._ravstack) - 1] += 1
        self._variation = "".join(
            _convert_integer_to_length_hex(i) for i in self.variationnumber
        )

    def modify_board_state(self, position_delta):
        """Delegate then modify board state and add index entries."""
        super().modify_board_state(position_delta)
        if len(self._ravstack) != len(self.halfmovenumber):
            while len(self._ravstack) < len(self.halfmovenumber):
                self.halfmovenumber.pop()
                self.variationnumber.pop()
            while len(self._ravstack) > len(self.halfmovenumber):
                self.halfmovenumber.append(self.halfmovenumber[-1])
                self.variationnumber.append(0)
            self._variation = "".join(
                _convert_integer_to_length_hex(i) for i in self.variationnumber
            )
        self.halfmovenumber[-1] += 1
        movenumber = _convert_integer_to_length_hex(self.halfmovenumber[-1])
        piecesquaremovekeys = self.piecesquaremovekeys
        piecemovekeys = self.piecemovekeys
        squaremovekeys = self.squaremovekeys
        pieces = [""] * 64
        bits = []
        mnv = movenumber + self._variation
        for piece in self._piece_placement_data.values():
            piece_name = piece.name
            piece_square = piece.square
            square_name = piece_square.name
            pieces[piece_square.number] = piece_name
            bits.append(piece_square.bit)

            # piecesquaremovekeys.append(mnv + piece_name + square_name)
            # squaremovekeys.append(mnv + mp[piece_name] + square_name)

            # If 'square piece' is better order than 'piece square'
            piecesquaremovekeys.append(mnv + square_name + piece_name)
            squaremovekeys.append(
                (
                    mnv
                    + square_name
                    + MAP_PGN_PIECE_TO_CQL_COMPOSITE_PIECE[piece_name]
                )
            )

        pieces = "".join(pieces)
        for piece_name in set(pieces):
            piecemovekeys.append(mnv + piece_name)
        delta_after = position_delta[1]
        self.positionkeys.append(
            sum(bits).to_bytes(8, "big").decode("iso-8859-1")
            + pieces
            + delta_after[1]
            + delta_after[3]
            + delta_after[2]
        )


class GameUpdateEstimate(GameUpdate):
    """Count characters and tokens to estimate time for import run."""

    start_char = 0
    end_char = 0

    def append_start_tag(self, match):
        """Set self.start_char if no PGN Tags found and delegate."""
        if not self._tags:
            self.start_char = match.start()
        super().append_start_tag(match)

    def append_game_termination_after_error(self, match):
        """Set self.end_char and delegate."""
        self.end_char = match.end()
        super().append_game_termination_after_error(match)

    def append_game_termination(self, match):
        """Set self.end_char and delegate."""
        self.end_char = match.end()
        super().append_game_termination(match)

    def append_bad_tag_and_set_error(self, match):
        """Set self.start_char if no PGN Tags found and delegate."""
        if not self._tags:
            self.start_char = match.start()
        super().append_bad_tag_and_set_error(match)


def _convert_integer_to_length_hex(integer):
    """Return hexadecimal representation of integer without '0x' prefix."""
    try:
        return MOVE_NUMBER_KEYS[integer]
    except IndexError:
        base16 = hex(integer)
        return str(len(base16) - 2) + base16[2:]



# options.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Functions for editing and applying font and colour options."""

import os

from . import constants
from . import fonts

font_names = {
    constants.MOVES_PLAYED_IN_GAME_FONT,
    constants.PIECES_ON_BOARD_FONT,
    constants.WILDPIECES_ON_BOARD_FONT,
    constants.LISTS_OF_GAMES_FONT,
    constants.TAGS_VARIATIONS_COMMENTS_FONT,
}

_OPTIONS_FILENAME = "options"


def get_saved_options(folder):
    """Return dictionary of colour and font defaults from options file.

    Return None if options file cannot be read or does not exist.

    """
    optionsfilename = os.path.join(folder, _OPTIONS_FILENAME)
    if not os.path.isfile(optionsfilename):
        return None
    try:
        with open(optionsfilename, "r") as optionsfile:
            return _extract_options(optionsfile)
    except OSError:
        pass
    return None


def save_options(folder, changes):
    """Save font and colour option changes in folder/<options file>.

    Changes are appended to the file.  The last occurrence of an option
    setting in the file is the one used when options file is read.

    A separate line is used for each option setting.  Format is:
    #<option>=<value>
    Leading and trailing whitespace is removed from <value> before use.

    """
    optionsfilename = os.path.join(folder, _OPTIONS_FILENAME)
    if os.path.exists(optionsfilename):
        if not os.path.isfile(optionsfilename):
            return
    with open(optionsfilename, "a+") as optionsfile:
        defaults = _extract_options(optionsfile)
        olddefaults, newdefaults = changes
        for key, value in olddefaults.items():
            if key in font_names:
                for vikey, vivalue in value.items():
                    if vivalue == newdefaults[key][vikey]:
                        del newdefaults[key][vikey]
                    elif defaults[key].get(vikey) == newdefaults[key][vikey]:
                        del newdefaults[key][vikey]
                if not newdefaults[key]:
                    del newdefaults[key]
            elif value == newdefaults[key]:
                del newdefaults[key]
            elif defaults.get(key) == newdefaults[key]:
                del newdefaults[key]
        newlines = []
        for key, value in newdefaults.items():
            if key in font_names:
                newlines.append("".join((key, "\n")))
                for vikey, vivalue in value.items():
                    if vikey in fonts.integer_attributes:
                        newlines.append(
                            "".join((vikey, "=", str(vivalue), "\n"))
                        )
                    else:
                        newlines.append("".join((vikey, "=", vivalue, "\n")))
            else:
                newlines.append("".join((key, "=", value, "\n")))
        if newlines:
            optionsfile.writelines(newlines)
    return


def _extract_options(fileid):
    """Extract options from fileid and return dictionary of options.

    The last occurrence of each option in the file is returned.

    """
    defaults = {
        constants.LITECOLOR_NAME: None,
        constants.DARKCOLOR_NAME: None,
        constants.WHITECOLOR_NAME: None,
        constants.BLACKCOLOR_NAME: None,
        constants.LINE_COLOR_NAME: None,
        constants.MOVE_COLOR_NAME: None,
        constants.ALTERNATIVE_MOVE_COLOR_NAME: None,
        constants.VARIATION_COLOR_NAME: None,
        constants.MOVES_PLAYED_IN_GAME_FONT: dict(),
        constants.PIECES_ON_BOARD_FONT: dict(),
        constants.WILDPIECES_ON_BOARD_FONT: dict(),
        constants.LISTS_OF_GAMES_FONT: dict(),
        constants.TAGS_VARIATIONS_COMMENTS_FONT: dict(),
    }
    for attr in fonts.modify_font_attributes:
        defaults[attr] = None
    font_details = False
    for line in fileid.readlines():
        text = line.strip()
        if text.startswith("#"):
            continue
        try:
            key_string, value_string = text.split("=", 1)
        except ValueError:
            key_string, value_string = text, ""
        key = key_string.strip()
        if key in defaults:
            if key in fonts.modify_font_attributes:
                if font_details:
                    defaults[font_details][key] = value_string.strip()
                continue
            if font_details:
                font_attributes = defaults[font_details]
                for attr in fonts.modify_font_attributes:
                    if defaults[attr]:
                        font_attributes[attr] = defaults[attr]
                        defaults[attr] = None
                font_details = False
            if key in font_names:
                font_details = key
            else:
                defaults[key] = value_string.strip()
    for attr in fonts.modify_font_attributes:
        del defaults[attr]
    return defaults



# enginedbedit.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit toplevel to edit or insert chess engine definition."""
import tkinter.messagebox

from solentware_grid.gui.dataedit import DataEdit

from .enginetoplevel import EngineToplevel, EngineToplevelEdit
from .topleveltext import EditText


class EngineDbEdit(EditText, DataEdit):
    """Edit chess engine definition on database, or insert a new record.

    parent is used as the master argument in EngineToplevelEdit calls.

    ui is used as the ui argument in EngineToplevelEdit calls.

    newobject, parent, oldobject, and the one or two EngineToplevelEdit
    instances created, are used as arguments in the super.__init__ call.

    showinitial determines whether a EngineToplevelEdit is created for
    oldobject if there is one.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "Engine Definition"

    def __init__(
        self,
        newobject=None,
        parent=None,
        oldobject=None,
        showinitial=True,
        ui=None,
    ):
        """Extend and create toplevel to edit or add chess engine definition.

        ui should be a UCI instance.

        """
        if not oldobject:
            showinitial = False
        super().__init__(
            newobject=newobject,
            parent=parent,
            oldobject=oldobject,
            newview=EngineToplevelEdit(master=parent, ui=ui),
            title="",
            oldview=EngineToplevel(master=parent, ui=ui)
            if showinitial
            else showinitial,
        )
        self._initialize()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a chess engine definition.

        Default value of object_ is oldobject attribute from DataEdit class.

        """
        if object_ is None:
            object_ = self.oldobject
        if object_:
            return "  ".join(
                (
                    self.text_name.join(("Edit ", ":")),
                    object_.value.get_name_text(),
                )
            )
        return "".join(("Insert ", self.text_name))

    @property
    def ui_base_table(self):
        """Return the User Interface EngineListGrid object."""
        return self.ui.base_engines

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.engines_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'editable' view."""
        return self.newview.ui

    @staticmethod
    def _set_item(view, object_):
        """Populate view with the engine definition extracted from object_."""
        view.definition.extract_engine_definition(object_.get_srvalue())
        view.set_engine_definition(object_.value)

    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        nedd = self.newview.get_name_engine_definition_dict()
        title = (self._get_title_for_object(),)
        if not nedd:
            tkinter.messagebox.showerror(
                parent=self.parent,
                title=title,
                message="".join(
                    (
                        "No chess engine definition given.\n\n",
                        "Name of chess engine definition must be ",
                        "first line, and subsequent lines the ",
                        "command to run the engine.",
                    )
                ),
            )
            return False
        self.newobject.value.load(repr(nedd))
        if not self.newobject.value.get_engine_command_text():
            tkinter.messagebox.showerror(
                parent=self.parent,
                title=title,
                message="".join(
                    (
                        "No chess engine definition given.\n\n",
                        "Name of chess engine definition must be ",
                        "first line, and subsequent lines the ",
                        "command to run the engine.",
                    )
                ),
            )
            return False
        url = self.newobject.value.engine_url_or_error_message()
        if isinstance(url, str):
            tkinter.messagebox.showerror(
                parent=self.parent, title=title, message=url
            )
            return False
        return super().dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after Toplevel destruction."""
        # ui_base_table is None when this happens other than directly closing
        # the Toplevel.
        try:
            super().tidy_on_destroy()
        except AttributeError:
            if self.ui_base_table is not None:
                raise



# __init__.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a NoSQL database of chess games using the vedis package.

This interface can be used if the vedis package is installed, from PyPI for
example.
"""



# engine.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess engine program definition."""
import os.path
from urllib.parse import urlsplit, parse_qs

from .constants import NAME_DELIMITER


class Engine:
    """Chess engine program definition.

    Maintain command line details for running a chess engine.

    The definition has a name: typically the chess engine name and version.

    The definition has a command line used to run the chess engine.

    """

    def __init__(self):
        """Delegate then initialize engine description and start command."""
        super().__init__()
        self._description_string = ""
        self._run_engine_string = ""

    def update_engine_definition(self, attributes):
        """Update existing self.__dict__ keys from dict attributes."""
        for attr in self.__dict__:
            if attr in attributes:
                self.__dict__[attr] = attributes[attr]

    def extract_engine_definition(self, text):
        """Return True if definition contains a command line and optional name.

        The command line starts with a path.  The last element of the path is
        used as the name if a separate name is not present.

        The definition contains at most two lines: the first line may be the
        optional name.

        """
        if isinstance(text, dict):
            self._description_string = text["_description_string"]
            self._run_engine_string = text["_run_engine_string"]
            return True
        definition = [t.strip() for t in text.split(NAME_DELIMITER)]
        if not definition[0] or not definition[-1]:
            return False
        if len(definition) > 2:
            return False
        self._run_engine_string = definition[-1]
        if len(definition) == 1:
            self._description_string = os.path.splitext(
                os.path.basename(definition[0].split()[0])
            )[0]
        else:
            self._description_string = definition[0]
        return True

    def get_name_text(self):
        """Return name text."""
        return self._description_string

    def get_name_engine_command_text(self):
        """Return name and command text."""
        return "\n".join(
            (
                self._description_string,
                self._run_engine_string,
            )
        )

    def get_engine_command_text(self):
        """Return command line to run engine."""
        return self._run_engine_string

    def _is_run_engine_command(self):
        """Return True if run engine command line starts with existing file."""
        if not self._run_engine_string:
            return False
        return os.path.isfile(self._run_engine_string.split()[0])

    def engine_url_or_error_message(self):
        """Return message string if invalid or urlsplit() object if valid."""
        url = urlsplit(self._run_engine_string)
        try:
            url.port
        except ValueError as exc:
            return "".join(
                (
                    "The port in the chess engine definition is ",
                    "invalid.\n\n",
                    "The reported error for the port is:\n\n",
                    str(exc),
                )
            )
        if not self._run_engine_string:
            return "".join(
                (
                    "The engine definition does not have a command to ",
                    "run chess engine.",
                )
            )
        if not (url.port or url.hostname):
            if not self._is_run_engine_command():
                return "".join(
                    (
                        "The engine definition command to run a chess engine ",
                        "does not name a file.",
                    )
                )
        if url.hostname or url.port:
            if url.path and url.query:
                return "".join(
                    (
                        "Engine must be query with hostname or port.\n\n",
                        "Path is: '",
                        url.path,
                        "'.\n\n",
                        "Query is: '",
                        url.query,
                        "'.\n",
                    )
                )
            if url.path:
                return "".join(
                    (
                        "Engine must be query with hostname or port.\n\n",
                        "Path is: '",
                        url.path,
                        "'.\n",
                    )
                )
            if not url.query:
                return "Engine must be query with hostname or port.\n\n"
            try:
                query = parse_qs(url.query, strict_parsing=True)
            except ValueError as exc:
                return "".join(
                    (
                        "Problem in chess engine specification.  ",
                        "The reported error is:\n\n'",
                        str(exc),
                        "'.\n",
                    )
                )
            if len(query) > 1:
                return "".join(
                    (
                        "Engine must be single 'key=value' or ",
                        "'value'.\n\n",
                        "Query is: '",
                        url.query,
                        "'.\n",
                    )
                )
            if len(query) == 1:
                for key, value in query.items():
                    if key != "name":
                        return "".join(
                            (
                                "Engine must be single 'key=value' or ",
                                "'value'.\n\n",
                                "Query is: '",
                                url.query,
                                "'\n\nand use ",
                                "'name' as key.\n",
                            )
                        )
                    if len(value) > 1:
                        return "".join(
                            (
                                "Engine must be single 'key=value' or ",
                                "'value'.\n\n",
                                "Query is: '",
                                url.query,
                                "' with more ",
                                "than one 'value'\n",
                            )
                        )
        elif url.path and url.query:
            return "".join(
                (
                    "Engine must be path without hostname or port.\n\n",
                    "Path is: '",
                    url.path,
                    "'.\n\n",
                    "Query is: '",
                    url.query,
                    "'.\n",
                )
            )
        elif url.query:
            return "".join(
                (
                    "Engine must be path without hostname or port.\n\n",
                    "Query is: '",
                    url.query,
                    "'.\n",
                )
            )
        elif not url.path:
            return "Engine must be path without hostname or port.\n"
        return url



# cqldisplay.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit Chess Query Language (ChessQL) statements.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

These four classes display ChessQL statements in the main window: they are used
in the cqlgrid module.

The _CQLDisplay class provides attributes and behaviour shared by the
CQLDisplay, CQLDisplayInsert, and CQLDisplayEdit, classes.  It also provides
properties to support implementation of behaviour shared with the Game*,
Repertoire*, and Query*, classes.

The CQLDisplay, CQLDisplayInsert, and CQLDisplayEdit, classes are subclasses
of the relevant ShowText, InsertText, EditText, DisplayText, and ListGamesText,
classes from the displaytext module; to implement behaviour shared with all
Text widgets in the main display (that includes widgets displaying PGN text).

"""

import tkinter
import tkinter.messagebox

from solentware_grid.gui.dataedit import RecordEdit
from solentware_grid.gui.datadelete import RecordDelete
from solentware_grid.core.dataclient import DataNotify

from solentware_bind.gui.bindings import Bindings

from .cql import CQL
from .cqledit import CQLEdit
from ..core.chessrecord import ChessDBrecordPartial
from ..core.cqlstatement import CQLStatement
from .eventspec import EventSpec
from .display import Display
from .displaytext import (
    ShowText,
    InsertText,
    EditText,
    DisplayText,
    ListGamesText,
)


class CQLDisplayListGamesError(Exception):
    """Exception class for display of lists of games."""


class _CQLDisplay(ShowText, DisplayText, CQL, Display, Bindings, DataNotify):
    """Extend and link ChessQL statement to database.

    sourceobject - link to database.

    Attribute binding_labels specifies the order navigation bindings appear
    in popup menus.

    Method _insert_item_database allows records to be inserted into a database
    from any CQL widget.

    """

    binding_labels = (
        EventSpec.navigate_to_position_grid,
        EventSpec.navigate_to_active_game,
        EventSpec.navigate_to_game_grid,
        EventSpec.navigate_to_repertoire_grid,
        EventSpec.navigate_to_active_repertoire,
        EventSpec.navigate_to_repertoire_game_grid,
        EventSpec.navigate_to_partial_grid,
        EventSpec.partialdisplay_to_previous_partial,
        EventSpec.partialdisplay_to_next_partial,
        EventSpec.navigate_to_partial_game_grid,
        EventSpec.navigate_to_selection_rule_grid,
        EventSpec.navigate_to_active_selection_rule,
        EventSpec.tab_traverse_backward,
        EventSpec.tab_traverse_forward,
    )

    def __init__(self, sourceobject=None, **ka):
        """Extend and link ChessQL statement to database."""
        super().__init__(**ka)
        self.blockchange = False
        if self.ui.base_partials.datasource:
            self.set_data_source(self.ui.base_partials.get_data_source())
        self.sourceobject = sourceobject
        self.insertonly = sourceobject is None
        self.recalculate_after_edit = sourceobject

    @property
    def ui_displayed_items(self):
        """Return manager of widgets displaying a partial position record."""
        return self.ui.partial_items

    @property
    def ui_configure_item_list_grid(self):
        """Return function to configure partial position grid to fit text."""
        return self.ui.configure_partial_grid

    @property
    def ui_set_item_name(self):
        """Return function to set status bar text to name of active query."""
        return self.ui.set_partial_name

    @property
    def ui_set_find_item_games(self):
        """Return function to set status bar text."""
        return self.ui.set_find_partial_name_games

    def _get_navigation_events(self):
        """Return event description tuple for navigation from query."""
        return (
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (EventSpec.partialdisplay_to_previous_partial, self._prior_item),
            (EventSpec.partialdisplay_to_next_partial, self._next_item),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (EventSpec.navigate_to_active_game, self.set_focus_gamepanel_item),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (EventSpec.export_from_partialdisplay, self._export_partial),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_round, self.traverse_round),
        )

    def delete_item_view(self, event=None):
        """Remove ChessQL statement item from screen."""
        del event
        self.ui.delete_position_view(self)

    def _insert_item_database(self, event=None):
        """Add ChessQL statement to database."""
        del event
        if self.ui.partial_items.active_item is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert ChessQL Statement",
                message="No active ChessQL statement to insert into database.",
            )
            return

        # This should see if ChessQL statement with same name already exists,
        # after checking for database open, and offer option to insert anyway.
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert ChessQL Statement",
                message="Cannot add ChessQL statement:\n\nNo database open.",
            )
            return

        datasource = self.ui.base_partials.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert ChessQL Statement",
                message="".join(
                    (
                        "Cannot add ChessQL statement:\n\n",
                        "Partial position list hidden.",
                    )
                ),
            )
            return
        updater = ChessDBrecordPartial()
        updater.value.process_statement(self.get_name_cql_statement_text())
        title = "Insert ChessQL Statement"
        tname = title.replace("Insert ", "").replace("S", "s")
        if not updater.value.get_name_text():
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title=title,
                message="".join(
                    (
                        "The '",
                        tname,
                        " has no name.\n\nPlease enter it's ",
                        "name as the first line of text.'",
                    )
                ),
            )
            return
        message = [
            "".join(
                (
                    "Confirm request to add ",
                    tname,
                    " named:\n\n",
                    updater.value.get_name_text(),
                    "\n\nto database.\n\n",
                )
            )
        ]
        if not updater.value.cql_error:
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title=title,
                message="".join(message),
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title=title,
                    message=tname.join(("Add ", " to database abandonned.")),
                )
                return
        else:
            message.append(updater.value.cql_error.get_error_report())
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title=title,
                message="".join(message),
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title=title,
                    message=tname.join(("Add ", " to database abandonned.")),
                )
                return
        editor = RecordEdit(updater, None)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        updater.key.recno = None  # 0
        editor.put()
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title=title,
            message="".join(
                (
                    tname,
                    ' "',
                    updater.value.get_name_text(),
                    '" added to database.',
                )
            ),
        )

    def on_game_change(self, instance):
        """Recalculate list of games for ChessQL statement after game update.

        instance is ignored: it is assumed a recalculation is needed.

        """
        del instance
        if self.sourceobject is not None:
            self._get_cql_statement_games_to_grid(
                self.cql_statement
            )  # .match)

    def generate_popup_navigation_maps(self):
        """Return genenal and current widget navigation binding maps."""
        navigation_map = dict(self._get_navigation_events())
        local_map = {}
        return navigation_map, local_map

    def _create_primary_activity_popup(self):
        """Delegate then add close command to popup and return popup menu."""
        popup = super()._create_primary_activity_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def on_partial_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if instance.newrecord:

            # Editing an existing record.
            value = instance.newrecord.value
            key = instance.newrecord.key

        else:

            # Inserting a new record or deleting an existing record.
            value = instance.value
            key = instance.key

        if self.sourceobject is not None:
            if (
                key == self.sourceobject.key
                and self.datasource.dbname == self.sourceobject.dbname
                and self.datasource.dbset == self.sourceobject.dbset
            ):
                self.blockchange = True

        pds = self.ui.partial_games.datasource
        if self.sourceobject is not None and key != self.sourceobject.key:
            pass
        elif self is not self.ui.partial_items.active_item:
            pass
        elif instance.newrecord is None:
            pds.forget_cql_statement_games(instance)
        elif instance.newrecord is False:
            try:
                pds.update_cql_statement_games(instance)
            except AttributeError as exc:
                if str(exc) == "'NoneType' object has no attribute 'answer'":
                    msg = "".join(
                        (
                            "Unable to add ChessQL statement to database, ",
                            "probably because an 'empty square' is in the ",
                            "query (eg '.a2-3'):\n\nThe reported  error ",
                            "is:\n\n",
                            str(exc),
                        )
                    )
                else:
                    msg = "".join(
                        (
                            "Unable to add ChessQL statement to database:\n\n",
                            "The reported error is:\n\n",
                            str(exc),
                        )
                    )
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="Insert ChessQL Statement",
                    message=msg,
                )
                return
        else:

            # Unfortunatly the existing list will have to be recalculated if
            # one of the caught exceptions occurs.
            pds.forget_cql_statement_games(instance)
            try:
                pds.update_cql_statement_games(instance.newrecord)
            except AttributeError as exc:
                if str(exc) == "'NoneType' object has no attribute 'answer'":
                    msg = "".join(
                        (
                            "Unable to edit ChessQL statement on database, ",
                            "probably because an 'empty square' is in the ",
                            "query (eg '.a2-3'):\n\nThe reported  error ",
                            "is:\n\n",
                            str(exc),
                        )
                    )
                else:
                    msg = "".join(
                        (
                            "Unable to edit ChessQL statement on database:",
                            "\n\nThe reported error is:\n\n",
                            str(exc),
                        )
                    )
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="Insert ChessQL Statement",
                    message=msg,
                )
                return

        if self is self.ui.partial_items.active_item:
            if self.sourceobject is not None and key == self.sourceobject.key:

                # Maybe should create a new CQLStatement instance.
                # self.cql_statement is a CQLStatement instance.
                # value is a ChessDBvaluePartial instance.
                self.cql_statement = value
                self.set_and_tag_item_text()
                self._get_cql_statement_games_to_grid(value)

    def get_text_for_statusbar(self):
        """Return 'Please wait ..' message for status bar."""
        return "".join(
            (
                "Please wait while finding games for ChessQL statement ",
                self.cql_statement.get_name_text(),
            )
        )

    def get_selection_text_for_statusbar(self):
        """Return CQL query name text for display in status bar."""
        return self.cql_statement.get_name_text()

    def set_game_list(self):
        """Delegate to refresh_game_list via 'after(...) call."""
        self.panel.after(
            0, func=self.try_command(self.ui.set_partial_name, self.panel)
        )
        self.panel.after(
            0, func=self.try_command(self.refresh_game_list, self.panel)
        )

    def _get_cql_statement_games_to_grid(self, statement):  # match):
        """Populate Partial Position games grid with games selected by match.

        "match" is named for the CQL version-1.0 keyword which started a CQL
        statement.  Usage is "(match ..." which become "cql(" no later than
        version 5.0 of CQL.  Thus "cql" is now a better name for the argument.

        """
        pgd = self.ui.partial_games
        if len(pgd.keys):
            key = pgd.keys[0]
        else:
            key = None
        pgd.close_client_cursor()
        try:
            pgd.datasource.get_cql_statement_games(
                statement, self.sourceobject
            )
        except AttributeError as exc:
            if str(exc) == "'NoneType' object has no attribute 'answer'":
                msg = "".join(
                    (
                        "Unable to list games for ChessQL statement, ",
                        "probably because an 'empty square' is in the query ",
                        "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                        str(exc),
                    )
                )
            else:
                msg = "".join(
                    (
                        "Unable to list games for ChessQL statement:\n\n",
                        "The reported error is:\n\n",
                        str(exc),
                    )
                )
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete ChessQL Statement",
                message=msg,
            )
            return
        pgd.fill_view(currentkey=key, exclude=False)
        if pgd.datasource.not_implemented:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="ChessQL Statement Not Implemented",
                message="".join(
                    (
                        "These filters are not implemented and ",
                        "are ignored:\n\n",
                        "\n".join(sorted(pgd.datasource.not_implemented)),
                    )
                ),
            )


class CQLDisplay(_CQLDisplay, CQL, DataNotify):
    """Display ChessQL statement from database and allow delete and insert.

    Method _delete_item_database allows records to be deleted from a database.

    """

    def _delete_item_database(self, event=None):
        """Remove ChessQL statement from database."""
        del event
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete ChessQL Statement",
                message="".join(
                    (
                        "Cannot delete ChessQL statement:\n\n",
                        "No database open.",
                    )
                ),
            )
            return
        datasource = self.ui.base_partials.get_data_source()
        if datasource is None:
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title="Delete ChessQL Statement",
                message="".join(
                    (
                        "Cannot delete ChessQL statement:\n\n",
                        "ChessQL statement list hidden.",
                    )
                ),
            )
            return
        if self.sourceobject is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete ChessQL Statement",
                message="".join(
                    (
                        "The ChessQL statement to delete has not ",
                        "been given.\n\nProbably because database ",
                        "has been closed and opened since this copy ",
                        "was displayed.",
                    )
                ),
            )
            return
        if self.blockchange:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete ChessQL Statement",
                message="\n".join(
                    (
                        "Cannot delete ChessQL statement.",
                        "Record has been amended since this copy displayed.",
                    )
                ),
            )
            return
        if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title="Delete ChessQL Statement",
            message="Confirm request to delete ChessQL statement.",
        ):
            return
        statement = self.cql_statement

        # Consider changing this since the call no longer ever returns None.
        if statement.is_statement() is not None:

            value = self.sourceobject.value
            if (
                statement.get_name_text() != value.get_name_text()
                or statement.is_statement() != value.is_statement()
                or statement.get_statement_text() != value.get_statement_text()
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="Delete ChessQL Statement",
                    message="\n".join(
                        (
                            "Cannot delete ChessQL statement.",
                            " ".join(
                                (
                                    "ChessQL statement on display is not",
                                    "same as rule from record.",
                                )
                            ),
                        )
                    ),
                )
                return
        editor = RecordDelete(self.sourceobject)
        editor.set_data_source(datasource, editor.on_data_change)
        editor.delete()
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title="Delete ChessQL Statement",
            message="".join(
                (
                    'ChessQL statement "',
                    self.sourceobject.value.get_name_text(),
                    '" deleted from database.',
                )
            ),
        )


class CQLDisplayInsert(
    ListGamesText, InsertText, _CQLDisplay, CQLEdit, DataNotify
):
    """Display ChessQL statement from database allowing insert.

    CQLEdit provides the widget and _CQLDisplay the database interface.

    Listing games for ChessQL statements is different to selection rule
    statements because selection rules share the listing area with the Game
    and Index options to the Select menu.  Index opens up the 'Move to' and
    Filter options too.  A definite user action is always required to generate
    game lists for selection rules, in addition to navigating to (giving focus
    to) the selection rule.  The area where ChessQL game lists are shown is
    dedicated to ChessQL statements.

    """

    def _get_list_games_events(self):
        """Return tuple of event bindings to list games for ChessQL rule."""
        return (
            (EventSpec.display_list, self._process_and_set_cql_statement_list),
        )

    def _add_list_games_entry_to_popup(self, popup):
        """Add option to list games for selection rule to popup."""
        # index argument added when change in method resolution order caused
        # this entry to be added after 'Close Item' rather than before.
        self._set_popup_bindings(
            popup, bindings=self._get_list_games_events(), index="Close Item"
        )

    def _set_database_navigation_close_item_bindings(self, switch=True):
        """Delegate then set list games event bindings."""
        super()._set_database_navigation_close_item_bindings(switch=switch)
        self.set_event_bindings_score(
            self._get_list_games_events(), switch=switch
        )

    def _process_and_set_cql_statement_list(self, event=None):
        """Display games with position matching edited ChessQL statement."""
        del event
        statement = CQLStatement()
        # Not sure this is needed or wanted.
        # statement.dbset = self.ui.base_games.datasource.dbset
        statement.process_statement(self.get_name_cql_statement_text())
        self.cql_statement = statement
        try:
            self.refresh_game_list()
        except AttributeError as exc:
            if str(exc) == "'NoneType' object has no attribute 'answer'":
                msg = "".join(
                    (
                        "Unable to list games for ChessQL statement, ",
                        "probably because an 'empty square' is in the query ",
                        "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                        str(exc),
                    )
                )
            else:
                msg = "".join(
                    (
                        "Unable to list games for ChessQL statement:\n\n",
                        "The reported error is:\n\n",
                        str(exc),
                    )
                )
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="ChessQL Statement",
                message=msg,
            )
        return "break"


class CQLDisplayEdit(EditText, CQLDisplayInsert):
    """Display ChessQL statement from database allowing edit and insert.

    Method _update_item_database allows records on the database to be amended.

    """

    def _update_item_database(self, event=None):
        """Modify existing ChessQL statement record."""
        del event
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit ChessQL Statement",
                message="Cannot edit ChessQL statement:\n\nNo database open.",
            )
            return
        datasource = self.ui.base_partials.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit ChessQL Statement",
                message="".join(
                    (
                        "Cannot edit ChessQL statement:\n\n",
                        "Partial position list hidden.",
                    )
                ),
            )
            return
        if self.sourceobject is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit ChessQL Statement",
                message="".join(
                    (
                        "The ChessQL statement to edit has not ",
                        "been given.\n\nProbably because database ",
                        "has been closed and opened since this copy ",
                        "was displayed.",
                    )
                ),
            )
            return
        if self.blockchange:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit ChessQL Statement",
                message="\n".join(
                    (
                        "Cannot edit ChessQL statement.",
                        "It has been amended since this copy was displayed.",
                    )
                ),
            )
            return
        original = ChessDBrecordPartial()
        original.load_record(
            (self.sourceobject.key.recno, self.sourceobject.srvalue)
        )

        # is it better to use DataClient directly?
        # Then original would not be used. Instead DataSource.new_row
        # gets record keyed by sourceobject and update is used to edit this.
        updater = ChessDBrecordPartial()
        updater.value.process_statement(self.get_name_cql_statement_text())
        title = "Edit ChessQL Statement"
        tname = title.replace("Edit ", "").replace("S", "s")
        if not updater.value.get_name_text():
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title=title,
                message="".join(
                    (
                        "The '",
                        tname,
                        " has no name.\n\nPlease enter it's ",
                        "name as the first line of text.'",
                    )
                ),
            )
            return
        message = [
            "".join(
                (
                    "Confirm request to edit ",
                    tname,
                    " named:\n\n",
                    updater.value.get_name_text(),
                    "\n\non database.\n\n",
                )
            )
        ]
        if not updater.value.cql_error:
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title=title,
                message="".join(message),
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title=title,
                    message=tname.join(("Edit ", " on database abandonned.")),
                )
                return
        else:
            message.append(updater.value.cql_error.get_error_report())
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title=title,
                message="".join(message),
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title=title,
                    message=tname.join(("Edit ", " on database abandonned.")),
                )
                return
        editor = RecordEdit(updater, original)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        original.set_database(editor.get_data_source().dbhome)
        updater.key.recno = original.key.recno
        editor.edit()
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title=title,
            message="".join(
                (
                    tname,
                    ' "',
                    updater.value.get_name_text(),
                    '" amended on database.',
                )
            ),
        )



# dbdudb.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The DbduDb class for berkeleydb and bsddb3 interfaces to Berkeley DB.

This module uses the solentware_base.core._db Database class.

"""

from solentware_base.core._db import Database

from .dptcompatdu import DptCompatdu


class DbduDb(DptCompatdu, Database):
    """Provide deferred update methods shared by Berkeley DB interfaces."""



# game_position.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide class for ChessDBrowGame and ChessDBrowPosition shared methods."""

from ..gui.gamedbedit import GameDbEdit
from ..gui.gamedbdelete import GameDbDelete
from ..gui.gamedbshow import GameDbShow


class GamePosition:
    """Provide methods shared by ChessDBrowGame and ChessDBrowPosition."""

    def show_row(self, dialog, oldobject):
        """Return a GameDbShow toplevel for oldobject record.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return GameDbShow(dialog, oldobject, ui=self.ui)

    def delete_row(self, dialog, oldobject):
        """Return a GameDbDelete toplevel for oldobject record.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return GameDbDelete(dialog, oldobject, ui=self.ui)

    def edit_row(self, dialog, newobject, oldobject, showinitial=True):
        """Return a GameDbEdit toplevel for newobject and oldobject records.

        dialog - a Toplevel
        newobject - a ChessDBrecordGame containing original data to be edited
        oldobject - a ChessDBrecordGame containing original data
        showintial == True - show both original and edited data

        """
        return GameDbEdit(
            newobject, dialog, oldobject, showinitial=showinitial, ui=self.ui
        )



# gamelistgrid.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of games on chess database."""

from solentware_grid.datagrid import DataGrid

from pgn_read.core.parser import PGN

from ..core.chessrecord import ChessDBrecordGameUpdate
from .gamedisplay import GameDisplay, GameDisplayEdit
from .constants import EMPTY_SEVEN_TAG_ROSTER
from ..core import export_game
from .display import Display
from ..shared.cql_gamelist_query import CQLGameListQuery
from ..shared.allgrid import AllGrid
from ..core.constants import UNKNOWN_RESULT


class _BookmarkStatusText:
    """Isolate set_selection_text call.

    This class is added to the ignored-classes list in pylint.conf.

    The associated super().bookmark_up and super().bookmark_down calls
    refer to DataGrid methods in the solentware_grid project.

    """

    def _set_selection_text_bookmark(self):
        """Show selection summary in status bar and hide call from pylint."""
        self.set_selection_text()


class GameListGrid(
    AllGrid,
    CQLGameListQuery,
    _BookmarkStatusText,
    DataGrid,
    Display,
):
    """A DataGrid for lists of chess games.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent, ui):
        """Extend with link to user interface object.

        parent - see superclass
        ui - container for user interface widgets and methods.

        """
        super().__init__(parent=parent)
        self.ui = ui
        self._configure_frame_and_initial_event_bindings()

    def _display_selected_item_kind(self, key, selected):
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplay is to be put.
        # Yes because GameDisplayEdit (see _edit_selected_item) includes
        # extra widgets. Want to say game.widget.destroy() eventually.
        # Read make_display_widget for GameDisplay and GameDisplayEdit.
        game = self.make_display_widget(selected)
        self.ui.add_game_to_display(game)
        self.ui.game_items.increment_object_count(key)
        self.ui.game_items.set_itemmap(game, key)
        self.ui.set_properties_on_all_game_grids(key)
        return game

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = GameDisplay(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        return game

    def _edit_selected_item(self, key):
        """Create display and return a GameDisplayEdit for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplayEdit is to be put.
        # Yes because GameDisplay (see _display_selected_item) includes
        # less widgets. Want to say game.widget.destroy() eventually.
        # Read make_edit_widget for GameDisplay and GameDisplayEdit.
        game = self.make_edit_widget(selected)
        self.ui.add_game_to_display(game)
        self.ui.game_items.increment_object_count(key)
        self.ui.game_items.set_itemmap(game, key)
        self.ui.set_properties_on_all_game_grids(key)
        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = GameDisplayEdit(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        return game

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for game key set or False."""
        if super().set_properties(key, dodefaultaction=False):
            return True
        if self.ui.game_items.object_display_count(key):
            self._set_background_on_display_row_under_pointer(key)
            return True
        if dodefaultaction:
            self._set_background_normal_row_under_pointer(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None."""
        row = super().set_row(key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.game_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        return None

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordGameUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key], oldobject, modal, title="Delete Game"
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordGameUpdate(),
            ChessDBrecordGameUpdate(),
            False,
            modal,
            title="Edit Game",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordGameUpdate(),
            ChessDBrecordGameUpdate(),
            True,
            modal,
            title="Edit Game",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordGameUpdate()
        instance = self.datasource.new_row()
        instance.srvalue = repr(EMPTY_SEVEN_TAG_ROSTER + UNKNOWN_RESULT)
        self.create_edit_dialog(
            instance, newobject, None, False, modal, title="New Game"
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordGameUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Game"
        )

    def _set_object_panel_item_properties(self):
        """Adjust properties of game_items to fit configure canvas event."""
        ui = self.ui
        for key in ui.game_items.object_panel_count:
            self.set_properties(key)

    @staticmethod
    def set_move_highlight(game):
        """Set move highlight at current position in game.

        In particular a game displayed from the list of games matching a
        position is shown at that position.

        """
        if game is not None:
            if game.current:
                game.set_move_highlight(game.current, True, True)

    def bookmark_down(self):
        """Extend to show selection summary in status bar."""
        super().bookmark_down()
        self._set_selection_text_bookmark()

    def bookmark_up(self):
        """Extend to show selection summary in status bar."""
        super().bookmark_up()
        self._set_selection_text_bookmark()

    def _export_text(self, event=None):
        """Export selected games as text."""
        del event
        self.ui.export_report(
            export_game.export_selected_games_text(
                self,
                self.ui.get_export_filename(
                    "Games (internal format)", pgn=False
                ),
            ),
            "Games (internal format)",
        )

    def export_pgn_import_format(self, event=None):
        """Export selected games in a PGN import format."""
        del event
        self.ui.export_report(
            export_game.export_selected_games_pgn_import_format(
                self,
                self.ui.get_export_filename("Games (import format)", pgn=True),
            ),
            "Games (import format)",
        )

    def _export_pgn(self, event=None):
        """Export selected games in PGN export format."""
        del event
        self.ui.export_report(
            export_game.export_selected_games_pgn(
                self, self.ui.get_export_filename("Games", pgn=True)
            ),
            "Games",
        )

    def _export_pgn_reduced_export_format(self, event=None):
        """Export selected games in PGN Reduced Export Format."""
        del event
        self.ui.export_report(
            export_game.export_selected_games_pgn_reduced_export_format(
                self,
                self.ui.get_export_filename(
                    "Games (reduced export format)", pgn=True
                ),
            ),
            "Games (reduced export format)",
        )

    def _export_pgn_no_comments_no_ravs(self, event=None):
        """Export selected games as PGN excluding all comments and RAVs."""
        del event
        self.ui.export_report(
            export_game.export_selected_games_pgn_no_comments_no_ravs(
                self,
                self.ui.get_export_filename(
                    "Games (no comments no ravs)", pgn=True
                ),
            ),
            "Games (no comments no ravs)",
        )

    def _export_pgn_no_comments(self, event=None):
        """Export selected games as PGN excluding all commentary tokens."""
        del event
        self.ui.export_report(
            export_game.export_selected_games_pgn_no_comments(
                self,
                self.ui.get_export_filename("Games (no comments)", pgn=True),
            ),
            "Games (no comments)",
        )



# export_game.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess game exporters."""

from pgn_read.core.parser import PGN

from . import chessrecord, filespec
from .export_pgn_import_format import get_game_pgn_import_format


def export_all_games_text(database, filename):
    """Export games in database to text file in internal record format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGameText()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.GAMES_FILE_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return True


def export_all_games_pgn(database, filename):
    """Export all database games in PGN export format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                if current_record[0] != prev_date:
                    for gfd in sorted(games_for_date):
                        gamesout.write(gfd[0])
                        gamesout.write("\n")
                        gamesout.write(gfd[2])
                        gamesout.write("\n")
                        gamesout.write(gfd[1])
                        gamesout.write("\n\n")
                    prev_date = current_record[0]
                    games_for_date = []
                game = database.get_primary_record(
                    filespec.GAMES_FILE_DEF, current_record[1]
                )
                try:
                    instance.load_record(game)
                except StopIteration:
                    break
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games_for_date.append(
                        instance.value.collected_game.get_export_pgn_elements()
                    )
                    if all_games_output is None:
                        all_games_output = True
                        no_games_output = False
                elif all_games_output:
                    if not no_games_output:
                        all_games_output = False
                current_record = cursor.next()
            for gfd in sorted(games_for_date):
                gamesout.write(gfd[0])
                gamesout.write("\n")
                gamesout.write(gfd[2])
                gamesout.write("\n")
                gamesout.write(gfd[1])
                gamesout.write("\n\n")
    finally:
        cursor.close()
    return all_games_output


def export_all_games_pgn_import_format(database, filename):
    """Export all database games in a PGN inport format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    all_games_output = None
    no_games_output = True
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.GAMES_FILE_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                try:
                    instance.load_record(current_record)
                except StopIteration:
                    break
                if instance.value.collected_game.is_pgn_valid_export_format():
                    gamesout.write(
                        get_game_pgn_import_format(
                            instance.value.collected_game
                        )
                    )
                    if all_games_output is None:
                        all_games_output = True
                        no_games_output = False
                elif all_games_output:
                    if not no_games_output:
                        all_games_output = False
                current_record = cursor.next()
    finally:
        cursor.close()
    return all_games_output


def export_all_games_pgn_no_comments(database, filename):
    """Export all database games in PGN export format excluding comments."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                if current_record[0] != prev_date:
                    for gfd in sorted(games_for_date):
                        gamesout.write(gfd[0])
                        gamesout.write("\n")
                        gamesout.write(gfd[2])
                        gamesout.write("\n")
                        gamesout.write(gfd[1])
                        gamesout.write("\n\n")
                    prev_date = current_record[0]
                    games_for_date = []
                game = database.get_primary_record(
                    filespec.GAMES_FILE_DEF, current_record[1]
                )
                try:
                    instance.load_record(game)
                except StopIteration:
                    break
                # Fix pycodestyle E501 (83 > 79 characters).
                # black formatting applied with line-length = 79.
                ivcg = instance.value.collected_game
                if ivcg.is_pgn_valid_export_format():
                    games_for_date.append(ivcg.get_export_pgn_rav_elements())
                    if all_games_output is None:
                        all_games_output = True
                        no_games_output = False
                elif all_games_output:
                    if not no_games_output:
                        all_games_output = False
                current_record = cursor.next()
            for gfd in sorted(games_for_date):
                gamesout.write(gfd[0])
                gamesout.write("\n")
                gamesout.write(gfd[2])
                gamesout.write("\n")
                gamesout.write(gfd[1])
                gamesout.write("\n\n")
    finally:
        cursor.close()
    return all_games_output


def export_all_games_pgn_no_comments_no_ravs(database, filename):
    """Export all database games, tags and moves only, in PGN export format.

    Comments and RAVs are excluded from the export.

    """
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                if current_record[0] != prev_date:
                    for gfd in sorted(games_for_date):
                        gamesout.write(gfd[0])
                        gamesout.write("\n")
                        gamesout.write(gfd[2])
                        gamesout.write("\n")
                        gamesout.write(gfd[1])
                        gamesout.write("\n\n")
                    prev_date = current_record[0]
                    games_for_date = []
                game = database.get_primary_record(
                    filespec.GAMES_FILE_DEF, current_record[1]
                )
                try:
                    instance.load_record(game)
                except StopIteration:
                    break
                collected_game = instance.value.collected_game
                if collected_game.is_pgn_valid_export_format():
                    games_for_date.append(
                        (
                            collected_game.get_seven_tag_roster_tags(),
                            collected_game.get_archive_movetext(),
                            collected_game.get_non_seven_tag_roster_tags(),
                        )
                    )
                    if all_games_output is None:
                        all_games_output = True
                        no_games_output = False
                elif all_games_output:
                    if not no_games_output:
                        all_games_output = False
                current_record = cursor.next()
            for gfd in sorted(games_for_date):
                gamesout.write(gfd[0])
                gamesout.write("\n")
                gamesout.write(gfd[2])
                gamesout.write("\n")
                gamesout.write(gfd[1])
                gamesout.write("\n\n")
    finally:
        cursor.close()
    return all_games_output


def export_all_games_pgn_reduced_export_format(database, filename):
    """Export all database games in PGN reduced export format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                if current_record[0] != prev_date:
                    for gfd in sorted(games_for_date):
                        gamesout.write(gfd[0])
                        gamesout.write("\n")
                        gamesout.write(gfd[1])
                        gamesout.write("\n\n")
                    prev_date = current_record[0]
                    games_for_date = []
                game = database.get_primary_record(
                    filespec.GAMES_FILE_DEF, current_record[1]
                )
                try:
                    instance.load_record(game)
                except StopIteration:
                    break
                # Fix pycodestyle E501 (80 > 79 characters).
                # black formatting applied with line-length = 79.
                ivcg = instance.value.collected_game
                if ivcg.is_pgn_valid_export_format():
                    games_for_date.append(ivcg.get_archive_pgn_elements())
                    if all_games_output is None:
                        all_games_output = True
                        no_games_output = False
                elif all_games_output:
                    if not no_games_output:
                        all_games_output = False
                current_record = cursor.next()
            for gfd in sorted(games_for_date):
                gamesout.write(gfd[0])
                gamesout.write("\n")
                gamesout.write(gfd[1])
                gamesout.write("\n\n")
    finally:
        cursor.close()
    return all_games_output


def export_selected_games_pgn_import_format(grid, filename):
    """Export selected records in a PGN import format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            if instance.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    get_game_pgn_import_format(instance.value.collected_game)
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(
                            instance.value.collected_game
                        )
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()

        # For all grids except ones displayed via 'Select | Rule | List Games'
        # the 'current_record = cursor.next()' can be immediately after the
        # 'while True:' statement, and the 'current_record = cursor.first()'
        # statement is redundant.
        # I think this implies a problem in the solentware_base RecordsetCursor
        # classes for each database engine since the 'finally:' clause should
        # kill the cursor.
        # The problem is only the first request outputs all the records to the
        # file.  Subsequent requests find no records to output, except that
        # doing some scrolling action resets the cursor and the next request
        # outputs all the records before the problem repeats.
        # The other methods in this class with this construct are affected too,
        # but this comment is not repeated.
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(
                            instance.value.collected_game
                        )
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()

    if len(games) == 0:
        return None
    with open(filename, "w") as gamesout:
        for game in games:
            gamesout.write(game)
            gamesout.write("\n\n")
    return all_games_output


def export_selected_games_pgn(grid, filename):
    """Export selected records in PGN export format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            if instance.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    instance.value.collected_game.get_export_pgn_elements()
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        instance.value.collected_game.get_export_pgn_elements()
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()

        # For all grids except ones displayed via 'Select | Rule | List Games'
        # the 'current_record = cursor.next()' can be immediately after the
        # 'while True:' statement, and the 'current_record = cursor.first()'
        # statement is redundant.
        # I think this implies a problem in the solentware_base RecordsetCursor
        # classes for each database engine since the 'finally:' clause should
        # kill the cursor.
        # The problem is only the first request outputs all the records to the
        # file.  Subsequent requests find no records to output, except that
        # doing some scrolling action resets the cursor and the next request
        # outputs all the records before the problem repeats.
        # The other methods in this class with this construct are affected too,
        # but this comment is not repeated.
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        instance.value.collected_game.get_export_pgn_elements()
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()

    if len(games) == 0:
        return None
    with open(filename, "w") as gamesout:
        for game in sorted(games):
            gamesout.write(game[0])
            gamesout.write("\n")
            gamesout.write(game[2])
            gamesout.write("\n")
            gamesout.write(game[1])
            gamesout.write("\n\n")
    return all_games_output


def export_selected_games_pgn_no_comments(grid, filename):
    """Export selected records in PGN export format excluding comments.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            if instance.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    instance.value.collected_game.get_export_pgn_rav_elements()
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                # Fix pycodestyle E501 (83 > 79 characters).
                # black formatting applied with line-length = 79.
                ivcg = instance.value.collected_game
                if ivcg.is_pgn_valid_export_format():
                    games.append(ivcg.get_export_pgn_rav_elements())
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                # Fix pycodestyle E501 (83 > 79 characters).
                # black formatting applied with line-length = 79.
                ivcg = instance.value.collected_game
                if ivcg.is_pgn_valid_export_format():
                    games.append(ivcg.get_export_pgn_rav_elements())
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    with open(filename, "w") as gamesout:
        for game in sorted(games):
            gamesout.write(game[0])
            gamesout.write("\n")
            gamesout.write(game[2])
            gamesout.write("\n")
            gamesout.write(game[1])
            gamesout.write("\n\n")
    return all_games_output


def export_selected_games_pgn_no_comments_no_ravs(grid, filename):
    """Export selected records in PGN export format excluding comments.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            collected_game = instance.value.collected_game
            if collected_game.is_pgn_valid_export_format():
                games.append(
                    (
                        collected_game.get_seven_tag_roster_tags(),
                        collected_game.get_archive_movetext(),
                        collected_game.get_non_seven_tag_roster_tags(),
                    )
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                collected_game = instance.value.collected_game
                if collected_game.is_pgn_valid_export_format():
                    games.append(
                        (
                            collected_game.get_seven_tag_roster_tags(),
                            collected_game.get_archive_movetext(),
                            collected_game.get_non_seven_tag_roster_tags(),
                        )
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                collected_game = instance.value.collected_game
                if collected_game.is_pgn_valid_export_format():
                    strt = collected_game.get_seven_tag_roster_tags()
                    nstrt = collected_game.get_non_seven_tag_roster_tags()
                    archive_movetext = collected_game.get_archive_movetext()
                    games.append((strt, archive_movetext, nstrt))
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    with open(filename, "w") as gamesout:
        for game in sorted(games):
            gamesout.write(game[0])
            gamesout.write("\n")
            gamesout.write(game[2])
            gamesout.write("\n")
            gamesout.write(game[1])
            gamesout.write("\n\n")
    return all_games_output


def export_selected_games_pgn_reduced_export_format(grid, filename):
    """Export selected records in grid to PGN file in reduced export format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            if instance.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    instance.value.collected_game.get_archive_pgn_elements()
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                # Fix pycodestyle E501 (80 > 79 characters).
                # black formatting applied with line-length = 79.
                ivcg = instance.value.collected_game
                if ivcg.is_pgn_valid_export_format():
                    games.append(ivcg.get_archive_pgn_elements())
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                # Fix pycodestyle E501 (80 > 79 characters).
                # black formatting applied with line-length = 79.
                ivcg = instance.value.collected_game
                if ivcg.is_pgn_valid_export_format():
                    games.append(ivcg.get_archive_pgn_elements())
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    with open(filename, "w") as gamesout:
        for game in sorted(games):
            gamesout.write(game[0])
            gamesout.write("\n")
            gamesout.write(game[1])
            gamesout.write("\n\n")
    return all_games_output


def export_selected_games_text(grid, filename):
    """Export selected records in grid to text file in internal record format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordGame()
    instance.set_database(database)
    games = []
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            games.append(instance.get_srvalue())
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                games.append(instance.get_srvalue())
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                games.append(instance.get_srvalue())
                current_record = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    with open(filename, "w") as gamesout:
        for game in games:
            gamesout.write(game)
            gamesout.write("\n")
    return True


def export_single_game_pgn_reduced_export_format(collected_game, filename):
    """Export collected_game to PGN file in reduced export format.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    with open(filename, "w") as gamesout:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_archive_movetext())
        gamesout.write("\n\n")


def export_single_game_pgn(collected_game, filename):
    """Export collected_game to filename in PGN export format.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    with open(filename, "w") as gamesout:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_non_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_all_movetext_in_pgn_export_format())
        gamesout.write("\n\n")


def export_single_game_pgn_no_comments_no_ravs(collected_game, filename):
    """Export collected_game tags and moves to filename in PGN export format.

    No comments or RAVs are included in the export (PGN Tags and moves
    played only).

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return None
    with open(filename, "w") as gamesout:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_non_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_archive_movetext())
        gamesout.write("\n\n")
    return True


def export_single_game_pgn_no_comments(collected_game, filename):
    """Export collected_game to filename in PGN export format without comments.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return None
    with open(filename, "w") as gamesout:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_non_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(
            collected_game.get_movetext_without_comments_in_pgn_export_format()
        )
        gamesout.write("\n\n")
    return True


def export_single_game_pgn_import_format(collected_game, filename):
    """Export collected_game to pgn file in a PGN import format."""
    if filename is None:
        return
    with open(filename, "w") as gamesout:
        gamesout.write(get_game_pgn_import_format(collected_game))


def export_single_game_text(collected_game, filename):
    """Export collected_game to text file in internal format."""
    if filename is None:
        return
    internal_format = next(PGN().read_games(collected_game.get_text_of_game()))
    with open(filename, "w") as gamesout:
        gamesout.write(internal_format.get_text_of_game())
        gamesout.write("\n")
